import numpy
import torch
from torch import nn
from config import config

class net(nn.Module):
	def __init__(self, in_dim, out_dim):
		super().__init__()

		self.encoder1 = nn.Linear(in_dim, 256)
		self.encoder2 = nn.Linear(256, 500)
		self.encoder3 = nn.Linear(256, 500)

		self.decoder1 = nn.Linear(500, 256)
		self.decoder2 = nn.Linear(256, out_dim)

	def encoder(self, x):
		h = torch.sigmoid(self.encoder1(x))
		mu = self.encoder2(h)
		log_var = self.encoder3(h)
		return mu, log_var # mean, variance

	def decoder(self, x):
		x = torch.sigmoid(self.decoder1(x))
		x = self.decoder2(x)
		return x

	def reparameterize(self, mu, log_var):
		eps = torch.randn(log_var.shape)
		std = torch.exp(log_var) ** 0.5
		return mu + std * eps

	def forward(self, x):
		mu, log_var = self.encoder(x)
		z = self.reparameterize(mu, log_var)
		x_hat = self.decoder(z)
		return x_hat, mu, log_var, z
