import pandas as pd
import numpy as np
import torch
from torch.utils import data
import VAE
import selection
from config import config

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

def minmaxscaler(data):
	min = np.min(data)
	max = np.max(data)	
	return (data - min) / (max - min)

class sampleData(data.Dataset):
	def __init__(self, X, y):
		self.X = X
		self.y = y
		
	def __getitem__(self, index):
		return self.X[index], self.y[index]
	
	def __len__(self):
		return len(self.X)

def AE(X, y, batch_size):
	dataset = sampleData(X, y)
	dataloader = data.DataLoader(dataset, batch_size = batch_size,
		shuffle = True, drop_last = False, **config.kwargs)
	netG = VAE.net(X.shape[1], X.shape[1]).to(device)
	# netG.load_state_dict(torch.load('./result/netG-01.pt'))
	netG.eval()
	for step, [X, y] in enumerate(dataloader):
		if (step == 0):
			x_hat, mu, log_var, hid = netG(X)
		else:
			x_hat, mu, log_var, h = netG(X)
			hid = torch.cat([hid, h], axis = 0)
	return hid

def construct(X, y):
	hid = AE(X, y, 128).detach().cpu().numpy()
	# hid = minmaxscaler(hid)
	index = selection.Ttest(hid, y, 250)
	return hid, index