import os
import numpy as np
import pandas as pd
import torch
from torch import nn
from torch import optim
from torch.utils import data
import matplotlib
import matplotlib.pyplot as plt
import GEV
import display
from config import config
import selection
from sklearn.svm import SVC
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.model_selection import KFold
from SMOTE import Smote
from predict_nn import predict_nn
from weight_nn import weight_nn
from dual_AE import dual_AE
from discriminator import discriminator
from generator import generator
from torch.autograd import Variable
from sklearn.metrics import roc_auc_score
import random


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
eps = 1e-8



class sampleData(data.Dataset):
    def __init__(self, FILE_NAME):
        Data = pd.read_csv(config.DATA_PATH + FILE_NAME, header = None).to_numpy()
        np.random.shuffle(Data)
        num=Data.shape[0]
        s1=Data[0][-1]
        for i in range(num):
            if(Data[i][-1]==s1):
                Data[i][-1]=1
            else:
                Data[i][-1]=0

        Pos = Data[Data[:, -1] == 1].astype(np.float32)
        Neg = Data[Data[:, -1] == 0].astype(np.float32)
        Pos_len= Pos.shape[0]
        Neg_len= Neg.shape[0]
        if(Pos_len<Neg_len):
        	g=Neg_len/Pos_len
        	g=int(g)
        	g=g-1
        	tmp=Pos[:,:-1]
        	tmp=Smote(tmp,N=g).over_sampling()
        	len1=tmp.shape[0]
        	len2=tmp.shape[1]
        	kk=np.zeros((len1,len2+1))
        	for i in range(len1):
        		for j in range(len2):
        			kk[i][j]=tmp[i][j]
        		kk[i][len2]=1
        	Pos=np.r_[Pos,kk]
        else:
            tmp=Neg[:,:-1]
            g=Pos_len/Neg_len
            g=int(g)
            g=g-1
            tmp=Smote(tmp,N=g).over_sampling()
            len1=tmp.shape[0]
            len2=tmp.shape[1]
            kk=np.zeros((len1,len2+1))
            for i in range(len1):
                for j in range(len2):
                    kk[i][j]=tmp[i][j]
                kk[i][len2]=0
            Neg=np.r_[Neg,kk]
        Pos = Pos.astype(np.float32)
        Neg = Neg.astype(np.float32)
        self.num = Pos.shape[0] + Neg.shape[0]
        np.random.shuffle(Pos)
        np.random.shuffle(Neg)
        d1 = int(Pos.shape[0] * 0.8)
        d2 = int(Neg.shape[0] * 0.8)
        self.test_data = np.r_[Pos[d1:, :-1], Neg[d2:, :-1]]
        self.test_label = np.r_[Pos[d1:, -1], Neg[d2:, -1]] 
        self.train_data = np.r_[Pos[:d1,:-1], Neg[:d2,:-1]]
        self.train_label = np.r_[Pos[:d1, -1], Neg[:d2, -1]]

        self.train_data = (self.train_data - self.train_data.mean(axis = 0, keepdims = True)) \
            / (self.train_data.std(axis = 0, keepdims = True) + eps)
        self.test_data = (self.test_data - self.test_data.mean(axis = 0, keepdims = True)) \
            / (self.test_data.std(axis = 0, keepdims = True) + eps)

    def __getitem__(self, index):
        return self.train_data[index], self.train_label[index]

    def __len__(self):
        return len(self.train_data)


def CEL(y, y_hat): #ensure y in {0, 1}
	return -(y * torch.log(y_hat + eps) + (1 - y) * torch.log(1 - y_hat + eps)).sum() / len(y)

def mean_regular(y,y_2,lam):
    ss=0
    for i in y_2:
        ss=ss+i
    ss=ss/len(y_2)
    squ_ss=(ss-0.5)*(ss-0.5)
    ans=lam*squ_ss*0.5
    return ans
def Loss(lambd, x_1, x_2, y_1, y_2, z_1,z_2):
	N = len(x_1)
	Loss_1 = nn.MSELoss(reduction = 'mean')
	v1 = Loss_1(x_1, x_2)
	Loss_2 = CEL
	v2 = Loss_2(y_1, y_2)
	v3 = Loss_1(z_1, z_2)
	regularzation=mean_regular(y_1,y_2,0.001)
	return 0.5*lambd * (v1+v3) + (1 - lambd) * v2+regularzation
	

def gradient_penalty(netD, real, fake):
	alpha = torch.rand(real.shape[0], 1).expand(real.shape).to(device)
	interpolates = alpha * real + (1 - alpha) * fake
	interpolates = torch.autograd.Variable(interpolates.to(device), requires_grad=True)
	disc_interpolates = netD(interpolates)
	grads = torch.autograd.grad(inputs = interpolates, outputs = disc_interpolates,
							grad_outputs = torch.ones(disc_interpolates.shape).to(device),
							create_graph = True, retain_graph = True,
							only_inputs = True)[0]
	gp = ((grads.norm(2, dim = 1) - 1) ** 2).mean()
	return gp

def GAN_CEL(x, x_hat):
	sm = nn.Softmax(dim = 1)
	x = sm(x)
	x_hat = sm(x)
	return -(x * torch.log(x_hat + eps) + (1 - x) * torch.log(1 - x_hat + eps)).sum() / x.nelement()

def main():

	FILE_LIST = os.listdir(config.DATA_PATH)
	FILE_LIST.sort(key = lambda x: x[:-4])
	file_num = 0
	mAcc = 0
	mG_mean = 0
	Dict = {}
	plt.figure(figsize = (60, 40))
	plt.subplots_adjust(wspace = 0.8, hspace = 0.8)
	times=0
	for i, FILE_NAME in enumerate(FILE_LIST):
		if FILE_NAME == '.DS_Store' or FILE_NAME == 'process.py' : continue
		print(i, FILE_NAME)
		dataset = sampleData(FILE_NAME)
		dataloader = data.DataLoader(dataset, batch_size = config.batch_size,
			shuffle = True, drop_last = False)
		X_train = torch.from_numpy(dataset.train_data)
		X_train=X_train.to(device)
		Y_train = torch.from_numpy(dataset.train_label)
		Y_train=Y_train.to(device)
		X_test = torch.from_numpy(dataset.test_data)
		X_test=X_test.to(device)
		Y_test = torch.from_numpy(dataset.test_label)
		Y_test=Y_test.to(device)
		file_num += 1
		sAcc = 0
		sG_mean = 0
		times = 0
		mauc=0
		mf1=0
		mp=0
		mr=0
		mMCC=0
		msen=0
		mspe=0
		for j in range(10):
			#model = GEV.Net(X_train.shape[1]).to(device)
			weight_=weight_nn(X_train.shape[1]).to(device)
			dual_=dual_AE(X_train.shape[1]).to(device)
			predict_=predict_nn(X_train.shape[1]).to(device)
			D = discriminator(X_train.shape[1]).to(device)

			criterion = Loss
			criterion_GAN=nn.BCELoss()

			optimizer_D = optim.Adam(D.parameters(), lr=config.alpha)
			optimizer_wei = optim.Adam(weight_.parameters(), lr = config.alpha)
			optimizer_dua = optim.Adam(dual_.parameters(), lr = config.alpha)
			optimizer_pre = optim.Adam(predict_.parameters(), lr = config.alpha)

			train_output = []
			test_output = []
			for epoch in range(config.GAN_epoch):
				for X,y in dataloader:
					X = X.to(device)
					y = y.to(device)
					####  the training of the discriminator
					x_2,z_dist,x_dist,z_out,z_1,z_2 = dual_(X)
					real_label = Variable(torch.ones((X.shape[0],1)))
					real_label = real_label.to(device)
					fake_label = Variable(torch.zeros((X.shape[0],1)))
					real_out = D(X)
					d_loss_real = criterion_GAN(real_out,real_label)
					#d_loss_real = GAN_CEL(real_out,real_label).mean()
					fake_label = fake_label.to(device)
					fake_out = D(x_2)
					d_loss_fake = criterion_GAN(fake_out,fake_label)
					#d_loss_fake = GAN_CEL(fake_out,fake_label).mean()
					gp = gradient_penalty(D, X, fake_out)
					d_loss = d_loss_real + d_loss_fake +10.*gp
					optimizer_D.zero_grad()
					d_loss.backward()
					optimizer_D.step()

					####   the training of the dual_AE
					x_2,z_dist,x_dist,z_out,z_1,z_2 = dual_(X)
					output = D(x_2)
					g_loss = criterion_GAN(output , real_label)
					optimizer_dua.zero_grad
					g_loss.backward()
					optimizer_dua.step()

			for epoch in range(config.n_epoch):
				if(epoch%20==0):
					print(epoch)
				for X, y in dataloader:
					X = X.to(device)
					y = y.to(device)
					x_1 = weight_(X)
					x_2,z_dist,x_dist,z_out,z_1,z_2 = dual_(X)
					FCL = torch.cat((x_1,z_dist,x_dist,z_out), 1)
					y_1 = predict_(FCL)
					#x_2, y_2,z_1,z_2, FCL = model(X)
					loss = criterion(config.lambd, X, x_2, y, y_1,z_1,z_2)
					optimizer_wei.zero_grad()
					optimizer_dua.zero_grad()
					optimizer_pre.zero_grad()

					loss.backward()

					optimizer_wei.step()
					optimizer_pre.step()
					optimizer_dua.step()

				#####
				# model.eval()
				# x_output_train, y_output_train,z_1,z_2, FCL = model(X_train)
				# train_output.append(criterion(config.lambd, X_train, x_output_train, Y_train, y_output_train,z_1,z_2).item())
				# x_output_test, y_output_test,z_1,z_2,FCL = model(X_test)
				# test_output.append(criterion(config.lambd, X_test, x_output_test, Y_test, y_output_test,z_1,z_2).item())
				# if (epoch > 25 and max(test_output[-20], test_output[-21], test_output[-22]) < min(test_output[-1], test_output[-2], test_output[-3])):
				# 	break

			x_1 = weight_(X_train)
			x_2,z_dist,x_dist,z_out,z_1,z_2 = dual_(X_train)
			FCL = torch.cat((x_1,z_dist,x_dist,z_out), 1)
			y_output = predict_(FCL)
			if (np.isnan(FCL.detach().cpu().numpy()).any()):
				continue


		####
			#Index = selection.Ttest(FCL.detach().cpu().numpy(), Y_train.detach().cpu().numpy(), 20)
		####


			x_1 = weight_(X_test)
			x_2,z_dist,x_dist,z_out,z_1,z_2 = dual_(X_test)
			FCL2 = torch.cat((x_1,z_dist,x_dist,z_out), 1)
			y_output = predict_(FCL2)



		###
			# y_output = SVC().fit((FCL.detach().cpu().numpy())[:, Index], Y_train.detach().cpu().numpy()
			# 	).predict((FCL2.detach().cpu().numpy())[:, Index])
			# y_output = torch.from_numpy(y_output)
			# y_output = y_output.to(device)
		###


			AUC=roc_auc_score(Y_test.detach().cpu().numpy(),y_output.detach().cpu().numpy())
			TP, TN, FP, FN, P, R, F1, Acc, G_mean,MCC,sen,spe = display.Eval(Y_test, y_output)

			print('Acc:%.4lf' % Acc, 'G_mean:%.4lf' % G_mean,'AUC:%.4lf'%AUC,'F1:%.4lf'%F1,
				'Precision:%.4lf'%P,'Recall:%.4lf'%R,'MCC:%.4lf'%MCC,'SEN:%.4lf'%sen,'SPE:%.4lf'%spe)
			if G_mean != 0:
				sAcc = max(sAcc, Acc)
				sG_mean = max(sG_mean, G_mean)

			mAcc += Acc
			mG_mean += G_mean
			mauc = mauc+AUC
			mf1 = mf1+F1
			mp+=P
			mr+=R
			mMCC= mMCC+MCC
			msen= msen+sen
			mspe= mspe+spe
			times=times+1


	print('mAcc:%.4lf' % (mAcc / times), 'mG_mean:%.4lf' % (mG_mean /times),'mAUC:%.4lf'%(mauc/times),'mF1:%.4lf'%(mf1/times),
		'mPrecision:%.4lf'%(mp/times),'mRecall:%.4lf'%(mr/times),'mMCC:%.4lf'%(mMCC/times),'msen:%.4lf'%(msen/times),
		'mspe:%.4lf'%(mspe/times),)


if (__name__=='__main__'):
	main()
